/********************************************************************
 *
 *                AUDIT - file auditor and sanity checker
 *
 *                by Jeff Tranter (tranter@pobox.com)
 *
 ********************************************************************
 *
 * Copyright (C) 1999 Jeff Tranter (tranter@pobox.com)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ********************************************************************
 *
 * See the man page for a description of what this program does and what
 * the requirements to run it are.
 *
 */


#include <assert.h>
#include <ctype.h>
#include <dirent.h>
#include <errno.h>
#ifdef GETOPTLONG
#include <getopt.h>
#endif /* GETOPTLONG */
#include <pwd.h>
#include <regex.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>

/* message severities */
typedef enum severity_t {INFO, WARN, ERROR, FATAL} severity_t;

/* global variables */
const char *version = "0.2 beta";  /* program version */
int verbose = 0;                   /* -v option */
char *user = 0;                    /* -u option */
int maxWarnings = 1000;            /* -m option */
int maxLevels = 100;               /* -l option */
int maxFileSize = 1000000;         /* --max-file-size option */
int minFileSize = 1;               /* --min-file-size option */
int maxAccessAge = 365;            /* --max-access-age option */
int maxModifyAge = 365;            /* --max-modify-age option */
int maxChangeAge = 365;            /* --max-change-age option */
int maxFilenameLength = 32;        /* --max-filename-length option */
int minFilenameLength = 2;         /* --min-filename-length option */
int maxLinks = 1;                  /* --max-links option */
int warnings = 0;                  /* number of warnings produced */
int errors = 0;                    /* number of errors produced */
int suppressed = 0;                /* number of errors suppressed */

/* Array listing which tests are suppressed. Index is test number. */
/* If value is non-zero, test is suppressed */
/* Somewhat wasteful of memory, but hey, it's only 8K */
char Ignore[8000];

/* Array of exception lines to ignore. Increase size if too small. */
char *Exception[10000];
/* Current index into exception array */
int exception = 0;

/* display command usage on standard error and exit */
static void Usage()
{
  fprintf(stderr,
	 "Audit version %s by Jeff Tranter (tranter@pobox.com)\n"
	 "Usage: audit [options] [<directory>...]\n"
	 "Options:\n"
#ifdef GETOPTLONG
	 "-h --help                     show command usage\n"
	 "-v --verbose                  verbose output\n"
	 "-u --user <username>          check from user's home directory and ownership\n"
	 "-m --max-warnings <n>         quit after <n> warning messages\n"
	 "-e --enable <#>               enable warning #\n"
	 "-s --suppress <#>             disable warning #\n"
	 "-l --max-levels <n>           maximum dir levels to search\n"
	 "-a --max-file-size <bytes>    flag files this size or larger\n"
	 "-b --min-file-size <bytes>    flags files this size or smaller\n"
	 "-c --max-access-age <days>    access time of files to flag as old\n"
	 "-d --max-modify-age <days>    modify time of files to flag as old\n"
	 "-g --max-change-age <days>    change time of files to flag as old\n"
	 "-h --max-filename-length <n>  flag file names >= this many chars\n"
	 "-i --min-filename-length <n>  flag file names <= this many chars\n"
	 "-j --max-links <n>            flag if file has more than <n> links\n", version);
#else
	 "-h                            show command usage\n"
	 "-v                            verbose output\n"
	 "-u <username>                 check from user's home directory and ownership\n"
	 "-m <n>                        quit after <n> warning messages\n"
	 "-e <#>                        enable warning #\n"
	 "-s <#>                        disable warning #\n"
	 "-l <n>                        maximum dir levels to search\n"
	 "-a <bytes>                    flag files this size or larger\n"
	 "-b <bytes>                    flags files this size or smaller\n"
	 "-c <days>                     access time of files to flag as old\n"
	 "-d <days>                     modify time of files to flag as old\n"
	 "-g <days>                     change time of files to flag as old\n"
	 "-h <n>                        flag file names >= this many chars\n"
	 "-i <n>                        flag file names <= this many chars\n"
	 "-j <n>                        flag if file has more than <n> links\n", version);
#endif /* GETOPTLONG */  

  exit(EXIT_FAILURE);
}


/* Parse command line options, setting global variables */
static int ParseOptions(int argc, char **argv, int fromRcFile)
{
const char *flags = "hvu:m:e:s:l:a:b:c:d:g:i:j:k:";
#ifdef GETOPTLONG
static struct option long_options[] =
{
  {"help",                no_argument,       NULL, 'h'},
  {"verbose",             no_argument,       NULL, 'v'},
  {"user",                required_argument, NULL, 'u'},
  {"max-warnings",        required_argument, NULL, 'm'},
  {"enable",              required_argument, NULL, 'e'},
  {"suppress",            required_argument, NULL, 's'},
  {"max-levels",          required_argument, NULL, 'l'},
  {"max-file-size",       required_argument, NULL, 'a'},
  {"min-file-size",       required_argument, NULL, 'b'},
  {"max-access-age",      required_argument, NULL, 'c'},
  {"max-modify-age",      required_argument, NULL, 'd'},
  {"max-change-age",      required_argument, NULL, 'g'},
  {"max-filename-length", required_argument, NULL, 'i'},
  {"min-filename-length", required_argument, NULL, 'j'},
  {"max-links",           required_argument, NULL, 'k'},
  {0, 0, 0, 0}
};
 int option_index;
#endif /* GETOPTLONG */
 int c;
 int test;
 
 /* resetting optind is needed when calling getopt more than once */
 optind = 1;
 
#ifdef GETOPTLONG
 while ((c = getopt_long(argc, argv, flags, long_options, &option_index)) != EOF) {
#else
   while ((c = getopt(argc, argv, flags)) != EOF) {
#endif /* GETOPTLONG */
     switch (c) {
     case 'h':
       Usage();
       break;
     case 'v':
       verbose = 1;
       break;
     case 'u':
       user = strdup(optarg);
       assert(user != 0);
       break;
     case 'm':
       maxWarnings = atoi(optarg);
       if (maxWarnings <= 0) {
	 fprintf(stderr, "audit: invalid argument to --max-warnings/-m option\n");
	 if (fromRcFile)
	   fprintf(stderr, "audit: error(s) found in .auditrc file, quitting\n");
	 exit(1);
       }
       break;
     case 'e':
       test = atoi(optarg);
       if (test <= 0 || test >= (int)(sizeof(Ignore)/sizeof(Ignore[0]))) {
	 fprintf(stderr, "audit: invalid argument to --enable/-e option\n");
	 if (fromRcFile)
	   fprintf(stderr, "audit: error(s) found in .auditrc file, quitting\n");
	 exit(1);
       }
       Ignore[test] = 0;
       break;
     case 's':
       test = atoi(optarg);
       if (test <= 0 || test >= (int)(sizeof(Ignore)/sizeof(Ignore[0]))) {
	 fprintf(stderr, "audit: invalid argument to --suppress/-s option\n");
	 exit(1);
       }
       Ignore[test] = 1;
       break;
     case 'l':
       maxLevels = atoi(optarg);
       if (maxLevels <= 0) {
	 fprintf(stderr, "audit: invalid argument to --max-levels/-l option\n");
	 if (fromRcFile)
	   fprintf(stderr, "audit: error(s) found in .auditrc file, quitting\n");
	 exit(1);
       }
       break;
     case 'a':
       maxFileSize = atoi(optarg);
       if (maxFileSize <= 0) {
	 fprintf(stderr, "audit: invalid argument to --max-file-size/-a option\n");
	 if (fromRcFile)
	   fprintf(stderr, "audit: error(s) found in .auditrc file, quitting\n");
	 exit(1);
       }
       break;
     case 'b':
       minFileSize = atoi(optarg);
       if (minFileSize <= 0) {
	 fprintf(stderr, "audit: invalid argument to --min-file-size/-a option\n");
	 if (fromRcFile)
	   fprintf(stderr, "audit: error(s) found in .auditrc file, quitting\n");
	 exit(1);
       }
       break;
     case 'c':
       maxAccessAge = atoi(optarg);
       if (maxAccessAge <= 0) {
	 fprintf(stderr, "audit: invalid argument to --max-access-age/-c option\n");
	 if (fromRcFile)
	   fprintf(stderr, "audit: error(s) found in .auditrc file, quitting\n");
	 exit(1);
       }
       break;
     case 'd':
       maxModifyAge = atoi(optarg);
       if (maxModifyAge <= 0) {
	 fprintf(stderr, "audit: invalid argument to --max-modify-age/-d option\n");
	 if (fromRcFile)
	   fprintf(stderr, "audit: error(s) found in .auditrc file, quitting\n");
	 exit(1);
       }
       break;
     case 'g':
       maxChangeAge = atoi(optarg);
       if (maxChangeAge <= 0) {
	 fprintf(stderr, "audit: invalid argument to --max-change-age/-d option\n");
	 if (fromRcFile)
	   fprintf(stderr, "audit: error(s) found in .auditrc file, quitting\n");
	 exit(1);
       }
       break;
     case 'i':
       maxFilenameLength = atoi(optarg);
       if (maxFilenameLength <= 0) {
	 fprintf(stderr, "audit: invalid argument to --max-filename-length/-i option\n");
	 if (fromRcFile)
	   fprintf(stderr, "audit: error(s) found in .auditrc file, quitting\n");
	 exit(1);
       }
       break;
     case 'j':
       minFilenameLength = atoi(optarg);
       if (minFilenameLength <= 0) {
	 fprintf(stderr, "audit: invalid argument to --min-filename-length/-j option\n");
	 if (fromRcFile)
	   fprintf(stderr, "audit: error(s) found in .auditrc file, quitting\n");
	 exit(1);
       }
       break;
     case 'k':
       maxLinks = atoi(optarg);
       if (maxLinks <= 0) {
	 fprintf(stderr, "audit: invalid argument to --max-links/-k option\n");
	 if (fromRcFile)
	   fprintf(stderr, "audit: error(s) found in .auditrc file, quitting\n");
	 exit(1);
       }
       break;
     case '?':
       if (fromRcFile)
	 fprintf(stderr, "audit: error(s) found in .auditrc file, quitting\n");
       exit(1);
       break;
     }
   }
   return optind;
 }

/*
Display an error message, e.g.
DisplayMessage(filename, WARN, 1234, "file size of %d bytes is greater than than maximum file size (%d bytes)", i, j);
audit: `/tmp/foo.bar': Warning 1234: file size of 0 bytes is less than minimum file size (2 bytes)
*/
static void DisplayMessage(const char *filename, int severity, int number, const char *format, ...)
{
  char msg[8192];
  char tmp[8192];
  char sev[16];
  int i, status;
  regex_t preg;
  va_list ap;

  /* skip if test has been suppressed */
  if ((severity == WARN) && (number > 0) && (number < (int)(sizeof(Ignore)/sizeof(Ignore[0]))) && Ignore[number])
    return;

  switch (severity) {
  case INFO:
    strncpy(sev, "Info", sizeof(sev));
    break;
  case WARN:
    strncpy(sev, "Warning", sizeof(sev));
    warnings++;
    break;
  case ERROR:
    strncpy(sev, "Error", sizeof(sev));
    errors++;
    break;
  case FATAL:
    strncpy(sev, "Fatal", sizeof(sev));
    break;
  default:
    assert(0);
    break;
  }
  if (filename != 0)
    sprintf(tmp, "audit: `%s': %s ", filename, sev);
  else
    sprintf(tmp, "audit: %s: ", sev);
  strncpy(msg, tmp, sizeof(msg));
  if (number != 0) {
    sprintf(tmp, "%d: ", number);
    strncat(msg, tmp, sizeof(msg));
  }

  va_start(ap, format);
  vsprintf(tmp, format, ap);
  va_end(ap);

  strncat(msg, tmp, sizeof(msg));
  if ((severity >= ERROR) && errno != 0) { /* add operating system error message, if any */
    strncat(msg, "\naudit: error was `", sizeof(msg));
    strncat(msg, strerror(errno), sizeof(msg));
    strncat(msg, "'", sizeof(msg));
    errno = 0 ;
  }

  /* Check for message which has been suppressed. */
  /* Can only suppress warnings. */
  /* If found, decrement the warning count */
  if (severity == WARN)
    for (i = 0; i < exception; i++) {
      status = regcomp(&preg, Exception[i], REG_EXTENDED);
      if (status != 0)
	fprintf(stderr, "audit: Fatal: compile of regular expression failed\n");
      if (regexec(&preg, msg, 0, 0, 0) == 0) {
	suppressed++;
	warnings--;
        regfree(&preg);
	return;
      }
      regfree(&preg);
    }
  
  /* print the message */
  fprintf(stderr, "%s\n", msg);

  if (severity == FATAL) {
    exit(1);
  }

  if (warnings >= maxWarnings) {
    fprintf(stderr, "audit: Fatal: maximum number of warnings reached, quitting\n");
    exit(1);
  }
}


/* return name of current user */
static char *Username()
{
  if (user != 0)
    return user;

  if (getlogin() != 0)
    return getlogin();
  if (getenv("LOGNAME") != 0)
    return getenv("LOGNAME");
  if (getenv("USER") != 0)
    return getenv("USER");
  DisplayMessage(0, FATAL, 0, "unable to get name of current user");
  return 0;
}


/* return home directory for user */
/* The value is cached in a static variable. */
static char *HomeDirectory(const char *userName)
{
  static struct passwd *passEntry = 0;

  if (passEntry == 0 || strcmp(userName, passEntry->pw_name)) {
    passEntry = getpwnam(userName);
    if (passEntry == 0)
      DisplayMessage(0, FATAL, 0, "unable to get password entry for user `%s'", userName);
  }
  return passEntry->pw_dir;
}


/* read data in .auditrc file, if any */
static void ParseRcFile()
{
  FILE *fp;
  char rcfile[1024];
  char line[1024];
  int argc = 0;
  char *argv[128];
  int i;

  argv[0] = "audit";
  argc++;

  /* construct filename */
  strncpy(rcfile, HomeDirectory(Username()), sizeof(rcfile));
  strncat(rcfile, "/.auditrc", sizeof(rcfile));

  /* quit if file does not exist */
  fp = fopen(rcfile, "r");
  if (fp == 0)
    return;

  /* read lines */
  while (fgets(line, sizeof(line), fp)) {

    /* strip off trailing newline */
    if (line[strlen(line)-1] == '\n')
      line[strlen(line)-1] = 0;

    /* check for blank line */
    if (strlen(line) == 0)
      continue;

    /* check for comment lines */
    if (line[0] == '#')
      continue;

    /* check for options */
    if (line[0] == '-') {
      if (argc >= (int)(sizeof(argv)/sizeof(argv[0]))) {
	errno = 0;
	DisplayMessage(0, FATAL, 0, "too many arguments in .auditrc file, quitting");
      }
      argv[argc] = strdup(line);
      assert(argv[argc] != 0);
      argc++;
      continue;
    }
    
    /* otherwise assume is exception */
    if (exception >= (int)(sizeof(Exception)/sizeof(Exception[0]))) {
      errno = 0;
      DisplayMessage(0, FATAL, 0, "too many exception lines in .auditrc file, quitting");
    }
    Exception[exception] = strdup(line);
    assert(Exception[exception] != 0);
    exception++;
  }

  /* parse options, if any */
  if (argc > 1)
    ParseOptions(argc, argv, 1);
  
  /* free up storage allocated for argv */
  for (i = 1; i <= argc; i++)
    free(argv[i]);

  /* and close file */
  fclose(fp);
}


/* Return base name of a file, e.g. given "/foo/bar/file.c" returns "file.c" */
/* Assumes Unix style pathnames (of course). */
static const char *BaseName(const char *path)
{
  if (strrchr(path, '/'))
    return strrchr(path, '/') + 1;
  else
    return path;
}


/* return true/false if string contains control characters */
static int HasControl(const char *name)
{
  while (*name) {
    if (iscntrl(*name))
      return 1;
    name++;
  }
  return 0;
}


/* return true/false if string contains non-ASCII characters */
static int HasNonAscii(const char *name)
{
  while (*name) {
    if (!isascii(*name))
      return 1;
    name++;
  }
  return 0;
}


/* return true/false if string contains white space characters */
static int HasSpace(const char *name)
{
  while (*name) {
    if (isspace(*name))
      return 1;
    name++;
  }
  return 0;
}


/* return true/false if string contains alphabetic characters */
static int HasAlpha(const char *name)
{
  while (*name) {
    if (isalpha(*name))
      return 1;
    name++;
  }
  return 0;
}


/* return true/false if string contains alphanumeric characters */
static int HasAlnum(const char *name)
{
  while (*name) {
    if (isalnum(*name))
      return 1;
    name++;
  }
  return 0;
}


static void PermissionChecks(const char *fileName, const struct stat *statbuf)
{
  /* permissions are not meaningful for symbolic links */
  if (S_ISLNK(statbuf->st_mode))
    return;
  if (statbuf->st_mode & 01000)
    DisplayMessage(fileName, WARN, 1001, "sticky bit set");
  if (statbuf->st_mode & 02000)
    DisplayMessage(fileName, WARN, 1002, "setgid bit set");
  if (statbuf->st_mode & 04000)
    DisplayMessage(fileName, WARN, 1003, "setuid bit set");
  if (statbuf->st_mode & 0002)
    DisplayMessage(fileName, WARN, 1004, "world writable");
  if (statbuf->st_mode & 0020)
    DisplayMessage(fileName, WARN, 1005, "group writable");
  if (!(statbuf->st_mode & 0400))
    DisplayMessage(fileName, WARN, 1006, "not readable by owner");
  if (!(statbuf->st_mode & 0040))
    DisplayMessage(fileName, WARN, 1007, "not readable by group");
  if (!(statbuf->st_mode & 0004))
    DisplayMessage(fileName, WARN, 1008, "not world readable");
  if ((statbuf->st_mode & 0040) && !(statbuf->st_mode & 0400))
    DisplayMessage(fileName, WARN, 1009, "readable by group but not owner");
  if ((statbuf->st_mode & 0004) && !(statbuf->st_mode & 0400))
    DisplayMessage(fileName, WARN, 1010, "readable by world but not owner");
  if ((statbuf->st_mode & 0010) && !(statbuf->st_mode & 0100))
    DisplayMessage(fileName, WARN, 1011, "executable by group but not owner");
  if ((statbuf->st_mode & 0001) && !(statbuf->st_mode & 0100))
    DisplayMessage(fileName, WARN, 1012, "executable by world but not owner");
  if ((statbuf->st_mode & 0100) && !(statbuf->st_mode & 0400))
    DisplayMessage(fileName, WARN, 1013, "executable but not readable by owner");
  if ((statbuf->st_mode & 0010) && !(statbuf->st_mode & 0040))
    DisplayMessage(fileName, WARN, 1014, "executable but not readable by group");
  if ((statbuf->st_mode & 0001) && !(statbuf->st_mode & 0004))
    DisplayMessage(fileName, WARN, 1015, "executable but not readable by world");
  if ((statbuf->st_mode & 04000) && !(statbuf->st_mode & 0100))
    DisplayMessage(fileName, WARN, 1016, "setuid but not execute");
  if ((statbuf->st_mode & 02000) && !(statbuf->st_mode & 0100))
    DisplayMessage(fileName, WARN, 1017, "setgid but not execute (mandatory file locking)");
  if (S_ISDIR(statbuf->st_mode) && !(statbuf->st_mode & 0100))
    DisplayMessage(fileName, WARN, 1018, "directory without search permission for user");
  if (S_ISDIR(statbuf->st_mode) && !(statbuf->st_mode & 0010))
    DisplayMessage(fileName, WARN, 1019, "directory without search permission for group");
  if (S_ISDIR(statbuf->st_mode) && !(statbuf->st_mode & 0001))
    DisplayMessage(fileName, WARN, 1020, "directory without search permission for world");
}


static void FileNameChecks(const char *fileName, const struct stat *statbuf)
{
  const char *name = BaseName(fileName);

  if ((int)strlen(name) < minFilenameLength)
    DisplayMessage(fileName, WARN, 2001, "short filename");
  if ((int)strlen(name) > maxFilenameLength)
    DisplayMessage(fileName, WARN, 2002, "long filename");
  if (HasControl(name))
    DisplayMessage(fileName, WARN, 2003, "filename contains control characters");
  if (HasNonAscii(name))
    DisplayMessage(fileName, WARN, 2004, "filename contains non-ASCII characters");
  if (HasSpace(name))
    DisplayMessage(fileName, WARN, 2005, "filename contains whitespace characters");
  if (!HasAlpha(name))
    DisplayMessage(fileName, WARN, 2006, "filename contains no alphabetic characters");
  if (!HasAlnum(name))
    DisplayMessage(fileName, WARN, 2007, "filename contains no alphanumeric characters");
  if (!strcmp(name, "core"))
    DisplayMessage(fileName, WARN, 2008, "possible core file");
  /* Ignore . files in home directory */
  if (name[0] == '.' && (strstr(fileName, HomeDirectory(user)) != fileName))
    DisplayMessage(fileName, WARN, 2009, "filename starts with `.'");
  if (name[strlen(name)-1] == '~')
    DisplayMessage(fileName, WARN, 2010, "filename ends in `~'");
  if ((name[0] == '#') && (name[strlen(name)-1] == '#'))
    DisplayMessage(fileName, WARN, 2011, "filename starts and ends in `#'");
  if ((strlen(name) >= 4) && !strcmp(&name[strlen(name)-4], ".bak"))
    DisplayMessage(fileName, WARN, 2012, "filename ends in `.bak'");
  if ((strlen(name) >= 4) && !strcmp(&name[strlen(name)-4], ".old"))
    DisplayMessage(fileName, WARN, 2013, "filename ends in `.old'");
  if ((strlen(name) >= 5) && !strcmp(&name[strlen(name)-5], ".orig"))
    DisplayMessage(fileName, WARN, 2014, "filename ends in `.orig'");
}


static void FileSizeChecks(const char *fileName, const struct stat *statbuf)
{
  if (statbuf->st_size == 0)
    DisplayMessage(fileName, WARN, 3001, "zero length file");
  if (statbuf->st_size > maxFileSize)
    DisplayMessage(fileName, WARN, 3002, "large file");
  if (statbuf->st_size < minFileSize)
    DisplayMessage(fileName, WARN, 3003, "small file");
}


static void FileTypeChecks(const char *fileName, const struct stat *statbuf)
{
  /* skip tests if filename starts with '/dev/' */
  if (strstr(fileName, "/dev/") == fileName)
      return;
  if (S_ISBLK(statbuf->st_mode))
    DisplayMessage(fileName, WARN, 4001, "block device");
  if (S_ISCHR(statbuf->st_mode))
    DisplayMessage(fileName, WARN, 4002, "character device");
  if (S_ISFIFO(statbuf->st_mode))
    DisplayMessage(fileName, WARN, 4003, "named pipe (FIFO)");
  if (S_ISSOCK(statbuf->st_mode))
    DisplayMessage(fileName, WARN, 4004, "socket");
  if (S_ISLNK(statbuf->st_mode))
    DisplayMessage(fileName, WARN, 4005, "symbolic link");
}


static void FileOwnershipChecks(const char *fileName, const struct stat *statbuf)
{
  struct passwd *passEntry = getpwnam(user);
  if (passEntry == 0)
    DisplayMessage(0, FATAL, 0, "unable to get password entry for user `%s'", user);
  if (statbuf->st_uid != passEntry->pw_uid)
    DisplayMessage(fileName, WARN, 5001, "owner is not user");
  if (statbuf->st_gid != passEntry->pw_gid)
    DisplayMessage(fileName, WARN, 5002, "group owner is not user's group");
}


static void LinkChecks(const char *fileName, const struct stat *statbuf)
{
  if (!S_ISDIR(statbuf->st_mode) && ((int)statbuf->st_nlink > maxLinks))
    DisplayMessage(fileName, WARN, 6001, "too many links", statbuf->st_nlink);
  if (S_ISDIR(statbuf->st_mode)) {
    struct dirent *dentry;
    DIR *dptr;
    int entries = 0;
    dptr = opendir(fileName);
    if (dptr == 0) {
      DisplayMessage(0, ERROR, 0, "opendir of `%s' failed, continuing", fileName);
      return;
    }
    while ((dentry = readdir(dptr)) != 0)
      entries++;
    if (entries <= 2)
      DisplayMessage(fileName, WARN, 6002, "empty directory");
    closedir(dptr);
  }
  /* Check for dangling symlink */
  if (S_ISLNK(statbuf->st_mode)) {
    struct stat statbuf;
    int status;
    status = stat(fileName, &statbuf);
    if (status == -1) {
      DisplayMessage(fileName, WARN, 6003, "dangling symbolic link");
    }
  }
}


static void DateStampChecks(const char *fileName, const struct stat *statbuf)
{
  time_t now = time(0);
  time_t old;
  if (now == -1)
    DisplayMessage(0, FATAL, 0, "unable to get current time");
  if (statbuf->st_atime > now)
    DisplayMessage(fileName, WARN, 7001, "access time in future");
  if (statbuf->st_mtime > now)
    DisplayMessage(fileName, WARN, 7002, "modify time in future");
  if (statbuf->st_ctime > now)
    DisplayMessage(fileName, WARN, 7003, "change time in future");
  old = now - (maxAccessAge * 86400);
  if (statbuf->st_atime < old)
    DisplayMessage(fileName, WARN, 7004, "access time in past");
  old = now - (maxModifyAge * 86400);
  if (statbuf->st_mtime < old)
    DisplayMessage(fileName, WARN, 7005, "modify time in past");
  old = now - (maxChangeAge * 86400);
  if (statbuf->st_ctime < old)
    DisplayMessage(fileName, WARN, 7006, "change time in past");
}


static void AuditFile(const char *fileName, const struct stat *statbuf)
{
  PermissionChecks(fileName, statbuf);
  FileNameChecks(fileName, statbuf);
  FileSizeChecks(fileName, statbuf);
  FileTypeChecks(fileName, statbuf);
  FileOwnershipChecks(fileName, statbuf);
  LinkChecks(fileName, statbuf);
  DateStampChecks(fileName, statbuf);
}


/* recursively audit a directory */
static void AuditDirectory(const char *dir)
{
  DIR *dptr;
  struct dirent *dentry;
  struct stat statbuf;
  int status;
  char *path;
  static int level;

  level++;
  if (level > maxLevels) {
    errno = 0;
    DisplayMessage(0, FATAL, 0, "maximum directory levels reached, quitting");
  }
  if (verbose) {
    DisplayMessage(0, INFO, 0, "auditing directory `%s'", dir);
  }
  dptr = opendir(dir);
  if (dptr == 0) {
    DisplayMessage(0, ERROR, 0, "opendir of `%s' failed, continuing", dir);
    return;
  }
  while ((dentry = readdir(dptr)) != 0) {
    /* skip '.' and '..' */
    if (!strcmp(dentry->d_name, ".") || !strcmp(dentry->d_name, ".."))
      continue;

    path = (char*) malloc(strlen(dir) + strlen(dentry->d_name) + 2);
    assert(path != 0);
    strcpy(path, dir);
    if (path[strlen(path)-1] != '/')
      strcat(path, "/");
    strcat(path, dentry->d_name);
    status = lstat(path, &statbuf);
    if (status == -1) {
      DisplayMessage(0, ERROR, 0, "stat of `%s' failed, continuing", path);
      return;
    }
    AuditFile(path, &statbuf);
    /* if directory, call self recursively */
    if (S_ISDIR(statbuf.st_mode))
      AuditDirectory(path);
    free(path);
  }
  closedir(dptr);
  level--;
}


/* Show all program settings. */
void ShowSettings()
{
  int i;

  fprintf(stderr,
	  "audit: Program version %s running with these settings:\n"
	  "       Using username `%s'\n"
	  "       Maximum warnings: %d\n"
	  "       Maximum directory levels: %d\n"
	  "       Maximum file size: %d chars\n"
	  "       Minimum file size: %d char(s)\n"
	  "       Maximum access time: %d days ago\n"
	  "       Maximum modify time: %d days ago\n"
	  "       Maximum change time: %d days ago\n"
	  "       Maximum filename length: %d chars\n"
	  "       Minimum filename length: %d chars\n"
	  "       Maximum number of file links: %d\n"
	  "       Exceptions in .auditrc file: %d\n"
	  "       Errors being suppressed: ",
	  version, user, maxWarnings, maxLevels, maxFileSize, minFileSize, maxAccessAge,
	  maxModifyAge, maxChangeAge, maxFilenameLength, minFilenameLength, maxLinks, exception);
  
  for (i = 0; i < (int)(sizeof(Ignore)/sizeof(Ignore[0])); i++) {
    if (Ignore[i] == 1)
      fprintf(stderr, "%d ", i);
  }
  fprintf(stderr, "\n");
}


int main(int argc, char **argv)
{
  int nextArg;
  int i;

  /* clear any errors */
  errno = 0;

  /* get options from .auditrc fle */
  ParseRcFile();

  /* get options from command line */
  nextArg = ParseOptions(argc, argv, 0);

  /* if not set by options, set user name to current user */
  if (user == 0)
    user = Username();

  if (verbose) {
    ShowSettings();
  }

  if (nextArg >= argc) {
    /* if no directory arguments, start at home directory */
    AuditDirectory(HomeDirectory(user));
  } else {
    /* otherwise use all directory arguments */
    for (i = nextArg; i < argc; i++)
      AuditDirectory(argv[i]);
  }

  if (verbose)
    DisplayMessage(0, INFO, 0, "completed with %d warnings, %d errors, %d suppressed errors",
		   warnings, errors, suppressed);

  exit(0);
}
